__version__ = "0.3.1"


import logging
import os

logging.basicConfig(level=os.getenv("LEVEL", "WARNING"))
logger = logging.getLogger(__name__)


class SensorWarning(UserWarning):
    """Recoverable errors"""

    pass


class WrongMessageFormat(SensorWarning):
    """Wrongly formattted message: throw away observation"""

    pass


class WrongMessageChecksum(SensorWarning):
    """Failed message checksum: throw away observation"""

    pass


class SensorWarmingUp(SensorWarning):
    """Empty message: throw away observation and wait until sensor warms up"""

    pass


class InconsistentObservation(SensorWarning):
    """Message payload makes no sense: throw away observation"""

    pass



from typer import Context, Option

from pms.service.influxdb import client_pub, influxdb
from pms.service.mqtt import client_sub, Data, mqtt


def bridge(
    mqtt_topic: str = Option("homie/+/+/+", help="mqtt root/topic"),
    mqtt_host: str = Option("mqtt.eclipse.org", help="mqtt server"),
    mqtt_port: int = Option(1883, help="server port"),
    mqtt_user: str = Option("", help="server username", show_default=False),
    mqtt_pass: str = Option("", help="server password", show_default=False),
    db_host: str = Option("influxdb", help="database server"),
    db_port: int = Option(8086, help="server port"),
    db_user: str = Option("root", help="server username"),
    db_pass: str = Option("root", help="server password"),
    db_name: str = Option("homie", help="database name"),
):
    """Bridge between MQTT and InfluxDB servers"""
    pub = client_pub(
        host=db_host, port=db_port, username=db_user, password=db_pass, db_name=db_name
    )

    def on_sensordata(data: Data) -> None:
        pub(time=data.time, tags={"location": data.location}, data={data.measurement: data.value})

    client_sub(
        topic=mqtt_topic,
        host=mqtt_host,
        port=mqtt_port,
        username=mqtt_user,
        password=mqtt_pass,
        on_sensordata=on_sensordata,
    )



from enum import Enum
from typing import Optional

from typer import Typer, Context, Option, echo, Exit

from pms import logger, __doc__, __version__
from pms.sensor import SensorReader
from pms.sensor.cli import serial, csv
from pms.service.cli import influxdb, mqtt, bridge


main = Typer(help=__doc__)
main.command()(serial)
main.command()(csv)
main.command()(influxdb)
main.command()(mqtt)
main.command()(bridge)


class Supported(str, Enum):
    PMSx003 = "PMSx003"
    PMS3003 = "PMS3003"
    PMS5003S = "PMS5003S"
    PMS5003ST = "PMS5003ST"
    PMS5003T = "PMS5003T"
    SDS01x = "SDS01x"
    SDS198 = "SDS198"
    HPMA115S0 = "HPMA115S0"
    HPMA115C0 = "HPMA115C0"
    SPS30 = "SPS30"
    MCU680 = "MCU680"
    default = PMSx003


def version_callback(value: bool):
    if value:
        name = __name__.split(".")[0]
        echo(f"{name} version {__version__}")
        raise Exit()


@main.callback()
def callback(
    ctx: Context,
    model: Supported = Option(Supported.default, "--sensor-model", "-m", help="sensor model"),
    port: str = Option("/dev/ttyUSB0", "--serial-port", "-s", help="serial port"),
    seconds: int = Option(60, "--interval", "-i", help="seconds to wait between updates"),
    samples: Optional[int] = Option(None, "--samples", "-n", help="stop after N samples"),
    debug: bool = Option(False, "--debug", help="print DEBUG/logging messages"),
    version: Optional[bool] = Option(None, "--version", callback=version_callback),
):
    """Read serial sensor"""
    if debug:
        logger.setLevel("DEBUG")
    ctx.obj = {"reader": SensorReader(model, port, seconds, samples)}



from enum import Enum
from pathlib import Path

from typing import Optional
from typer import Context, Option, echo

from pms import logger


class Format(str, Enum):
    csv = "csv"
    pm = "pm"
    num = "num"
    raw = "raw"
    cf = "cf"
    atm = "atm"
    hcho = "hcho"
    bme = "bme"
    bsec = "bsec"


def serial(
    ctx: Context,
    format: Optional[Format] = Option(None, "--format", "-f", help="formatted output"),
):
    """Read sensor and print measurements"""
    with ctx.obj["reader"] as reader:
        if format:
            if format == "csv":
                obs = next(reader())
                echo(f"{obs:header}")
            for obs in reader():
                echo(f"{obs:{format}}")
        else:
            for obs in reader():
                echo(str(obs))


def csv(
    ctx: Context,
    path: Path = Option(Path("pms.csv"), "--filename", "-F", help="csv formatted file"),
    overwrite: bool = Option(False, "--overwrite", help="overwrite file, if already exists"),
):
    """Read sensor and print measurements"""
    mode = "w" if overwrite else "a"
    logger.debug(f"open {path} on '{mode}' mode")
    with ctx.obj["reader"] as reader, path.open(mode) as f:
        # add header to new files
        if path.stat().st_size == 0:
            obs = next(reader())
            f.write(f"{obs:header}\n")
        for obs in reader():
            f.write(f"{obs:csv}\n")



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, asdict
from typing import NamedTuple, Tuple, Dict
from datetime import datetime
from pms import logger, WrongMessageFormat


class Cmd(NamedTuple):
    """Single command"""

    command: bytes
    answer_header: bytes
    answer_length: int


class Commands(NamedTuple):
    """Required commands"""

    passive_read: Cmd
    passive_mode: Cmd
    active_mode: Cmd
    sleep: Cmd
    wake: Cmd


class Message(metaclass=ABCMeta):
    """
    Base class for serial messages from PM sensors
    """

    def __init__(self, message: bytes) -> None:
        logger.debug(f"message hex: {message.hex()}")
        self.message = message

    @classmethod
    def unpack(cls, message: bytes, header: bytes, length: int) -> Tuple[float, ...]:
        try:
            # validate full message
            msg = cls._validate(message, header, length)
        except WrongMessageFormat as e:
            # search last complete message on buffer
            start = message.rfind(header, 0, 4 - length)
            if start < 0:  # No match found
                raise
            # validate last complete message
            msg = cls._validate(message[start : start + length], header, length)

        # data: unpacked payload
        payload = cls._unpack(msg.payload)
        logger.debug(f"message payload: {payload}")
        return payload

    @classmethod
    def decode(cls, message: bytes, command: Cmd) -> Tuple[float, ...]:
        header = command.answer_header
        length = command.answer_length
        return cls.unpack(message, header, length)[cls.data_records]  # type: ignore

    @property
    @classmethod
    @abstractmethod
    def data_records(cls) -> slice:  # pragma: no cover
        pass

    @property
    @abstractmethod
    def header(self) -> bytes:  # pragma: no cover
        pass

    @property
    @abstractmethod
    def payload(self) -> bytes:  # pragma: no cover
        pass

    @property
    @abstractmethod
    def checksum(self) -> int:  # pragma: no cover
        pass

    @classmethod
    @abstractmethod
    def _validate(
        self, message: bytes, header: bytes, length: int
    ) -> "Message":  # pragma: no cover
        pass

    @staticmethod
    @abstractmethod
    def _unpack(message: bytes) -> Tuple[float, ...]:  # pragma: no cover
        pass


@dataclass  # type: ignore
class ObsData(metaclass=ABCMeta):
    """Measurements

    time: measurement time [seconds since epoch]
    date: measurement time [datetime object]
    """

    time: int

    @property
    def date(self) -> datetime:
        """measurement time as datetime object"""
        return datetime.fromtimestamp(self.time)

    def subset(self, spec: str) -> Dict[str, float]:  # pragma: no cover
        logger.warning("The 'subset' is deprecated", DeprecationWarning, stacklevel=2)
        if spec:
            obs = {k: v for k, v in asdict(self).items() if k.startswith(spec)}
        else:
            obs = {k: v for k, v in asdict(self).items() if k != "time"}
        if obs:
            return obs
        raise ValueError(  # pragma: no cover
            f"Unknown subset code '{spec}' "
            f"for object of type '{__name__}.{self.__class__.__name__}'"
        )

    @abstractmethod
    def __format__(self, spec: str) -> str:
        if spec == "header":  # header for csv file
            return ", ".join(asdict(self).keys())
        raise ValueError(  # pragma: no cover
            f"Unknown format code '{spec}' "
            f"for object of type '{__name__}.{self.__class__.__name__}'"
        )

    def __str__(self):
        return self.__format__("pm")


def metadata(long_name: str, units: str, topic: str):
    """For fields(metadata=metadata(...))"""
    return dict(long_name=long_name, units=units, topic=topic)



"""
Senserion SPS30 sensors
- message protocol implements byte-stuffing
- there is no active mode read
- passive read messages are 47b long
- empty read messages are 7b long
"""

from dataclasses import dataclass, field
from typing import Tuple
import struct

from pms import WrongMessageFormat, WrongMessageChecksum, SensorWarmingUp
from pms.sensor import base

commands = base.Commands(
    # Read Measured Values
    passive_read=base.Cmd(b"\x7E\x00\x03\x00\xFC\x7E", b"\x7E\x00\x03\x00\x28", 47),
    passive_mode=base.Cmd(b"", b"", 0),
    active_mode=base.Cmd(b"", b"", 0),
    # Stop Measurement
    sleep=base.Cmd(b"\x7E\x00\x01\x00\xFE\x7E", b"\x7E\x00\x01\x00\x00", 7),
    # Start Measurement
    wake=base.Cmd(b"\x7E\x00\x00\x02\x01\x03\xF9\x7E", b"\x7E\x00\x00", 7),
)


class Message(base.Message):
    """Messages from Senserion SPS30 sensors"""

    data_records = slice(10)

    @property
    def header(self) -> bytes:
        return self.message[:5]

    @property
    def payload(self) -> bytes:
        """byte de-stuffing"""
        p = self.message[5:-2]
        for k, v in {
            b"\x7E": b"\x7D\x5E",
            b"\x7D": b"\x7D\x5D",
            b"\x11": b"\x7D\x31",
            b"\x13": b"\x7D\x33",
        }.items():
            if v in p:  # pragma: no cover
                p = p.replace(v, k)
        return p

    @property
    def checksum(self) -> int:
        return self.message[-2]

    @property
    def tail(self) -> int:
        return self.message[-1]

    @classmethod
    def _validate(cls, message: bytes, header: bytes, length: int) -> base.Message:

        # consistency check: bug in message singnature
        assert len(header) == 5, f"wrong header length {len(header)}"
        assert header[:2] == b"\x7E\x00", f"wrong header start {header!r}"
        assert length in [7, 47], f"wrong payload length {length} != 4||47"
        len_payload = header[-1]
        assert length == len_payload + 7, f"wrong payload length {length} != {len_payload+7}"

        # validate message: recoverable errors (throw away observation)
        msg = cls(message)
        if msg.header != header:
            raise WrongMessageFormat(f"message header: {msg.header!r}")
        if msg.tail != 0x7E:
            raise WrongMessageFormat(f"message tail: {msg.tail:#x}")
        if len(message) != length:
            raise WrongMessageFormat(f"message length: {len(message)} != {length}")
        checksum = 0xFF - (sum(msg.header[1:]) + sum(msg.payload)) % 0x100
        if msg.checksum != checksum:
            raise WrongMessageChecksum(f"message checksum {msg.checksum} != {checksum}")
        if sum(msg.payload) == 0:
            raise SensorWarmingUp(f"message empty: warming up sensor")
        return msg

    @staticmethod
    def _unpack(message: bytes) -> Tuple[float, ...]:
        return struct.unpack(f">{len(message)//4}f", message)


@dataclass(frozen=False)
class ObsData(base.ObsData):
    """SPS30 observations

    time                                    measurement time [seconds since epoch]
    pm01, pm25, pm04, pm10                  PM1.0, PM2.5, PM4.0, PM10 [ug/m3]
    n0_5, n1_0, n2_5, n4_0, n10_0           number concentrations between 0.3 and X.Y um [#/cm3]
    diam                                    typical particle size [μm]
    """

    pm01: float = field(metadata=base.metadata("PM1", "ug/m3", "concentration"))
    pm25: float = field(metadata=base.metadata("PM2.5", "ug/m3", "concentration"))
    pm04: float = field(metadata=base.metadata("PM4", "ug/m3", "concentration"))
    pm10: float = field(metadata=base.metadata("PM10", "ug/m3", "concentration"))
    n0_5: float
    n1_0: float
    n2_5: float
    n4_0: float
    n10_0: float
    diam: float

    def __format__(self, spec: str) -> str:
        if spec == "header":
            return super().__format__(spec)
        if spec == "pm":
            return f"{self.date:%F %T}: PM1 {self.pm01:.1f}, PM2.5 {self.pm25:.1f}, PM4 {self.pm04:.1f}, PM10 {self.pm10:.1f} ug/m3"
        if spec == "csv":
            pm = f"{self.pm01:.1f}, {self.pm25:.1f}, {self.pm04:.1f}, {self.pm10:.1f}"
            num = f"{self.n0_5:.2f}, {self.n1_0:.2f}, {self.n2_5:.2f}, {self.n4_0:.2f}, {self.n10_0:.2f}"
            return f"{self.time}, {pm}, {num}, {self.diam:.1f}"
        if spec == "num":
            return f"{self.date:%F %T}: N0.5 {self.n0_5:.2f}, N1.0 {self.n1_0:.2f}, N2.5 {self.n2_5:.2f}, N4.0 {self.n4_0:.2f}, N10 {self.n10_0:.2f} #/cm3"
        if spec == "diam":
            return f"{self.date:%F %T}: Ø {self.diam:.1f} μm"
        raise ValueError(  # pragma: no cover
            f"Unknown format code '{spec}' "
            f"for object of type '{__name__}.{self.__class__.__name__}'"
        )



from datetime import datetime
from dataclasses import fields
from typing import Dict, Union, Callable, NamedTuple

from typer import Context, Option, style, colors, echo, Abort

try:
    from paho.mqtt import client
except ModuleNotFoundError:  # pragma: no cover
    client = None  # type: ignore

from pms import logger
from pms.sensor.base import ObsData


def __missing_mqtt():  # pragma: no cover
    name = style(__name__, fg=colors.GREEN, bold=True)
    package = style("pypms", fg=colors.GREEN, bold=True)
    module = style("paho-mqtt", fg=colors.RED, bold=True)
    extra = style("mqtt", fg=colors.RED, bold=True)
    pip = style("python3 -m pip instal --upgrade", fg=colors.GREEN)
    pipx = style("pipx inject", fg=colors.GREEN)
    echo(
        f"""
{name} provides additional functionality to {package}.
This functionality requires the {module} module, which is not installed.
You can install this additional dependency with
\t{pip} {package}[{extra}]
Or, if you installed {package} with pipx
\t{pipx} {package} {module}
"""
    )
    raise Abort()


def client_pub(
    *, topic: str, host: str, port: int, username: str, password: str
) -> Callable[[Dict[str, Union[int, str]]], None]:  # pragma: no cover
    if client is None:
        __missing_mqtt()
    c = client.Client(topic)
    c.enable_logger(logger)
    if username:
        c.username_pw_set(username, password)

    c.on_connect = lambda client, userdata, flags, rc: client.publish(
        f"{topic}/$online", "true", 1, True
    )
    c.will_set(f"{topic}/$online", "false", 1, True)
    c.connect(host, port, 60)
    c.loop_start()

    def pub(data: Dict[str, Union[int, str]]) -> None:
        for k, v in data.items():
            c.publish(f"{topic}/{k}", v, 1, True)

    return pub


class Data(NamedTuple):
    time: int
    location: str
    measurement: str
    value: float

    @staticmethod
    def now() -> int:  # pragma: no cover
        """current time as seconds since epoch"""
        return int(datetime.now().timestamp())

    @classmethod
    def decode(cls, topic: str, payload: str, *, time: int = None) -> "Data":
        """Decode a MQTT message

        For example
        >>> decode("homie/test/pm10/concentration", "27")
        >>> Data(now(), "test", "pm10", 27)
        """
        if not time:  # pragma: no cover
            time = cls.now()

        fields = topic.split("/")
        if len(fields) != 4:
            raise UserWarning(f"topic total length: {len(fields)}")
        if any([f.startswith("$") for f in fields]):
            raise UserWarning(f"system topic: {topic}")
        location, measurement = fields[1:3]

        try:
            value = float(payload)
        except ValueError:  # pragma: no cover
            raise UserWarning(f"non numeric payload: {payload}")
        else:
            return cls(time, location, measurement, value)


def client_sub(
    topic: str,
    host: str,
    port: int,
    username: str,
    password: str,
    *,
    on_sensordata: Callable[[Data], None],
) -> None:  # pragma: no cover
    def on_message(client, userdata, msg):
        try:
            data = Data.decode(msg.topic, msg.payload)
        except UserWarning as e:
            logger.debug(e)
        else:
            on_sensordata(data)

    if client is None:
        __missing_mqtt()
    c = client.Client(topic)
    c.enable_logger(logger)
    if username:
        c.username_pw_set(username, password)

    c.on_connect = lambda client, userdata, flags, rc: client.subscribe(topic)
    c.on_message = on_message
    c.connect(host, port, 60)
    c.loop_forever()


def mqtt(
    ctx: Context,
    topic: str = Option("homie/test", "--topic", "-t", help="mqtt root/topic"),
    host: str = Option("mqtt.eclipse.org", "--mqtt-host", help="mqtt server"),
    port: int = Option(1883, "--mqtt-port", help="server port"),
    user: str = Option("", "--mqtt-user", help="server username", show_default=False),
    word: str = Option("", "--mqtt-pass", help="server password", show_default=False),
):
    """Read sensor and push PM measurements to a MQTT server"""
    pub = client_pub(topic=topic, host=host, port=port, username=user, password=word)

    def publish(obs: ObsData, metadata: bool = False):
        data = {}
        for field in fields(obs):
            if not field.metadata:
                continue
            if metadata:
                data[f"{field.name}/$type"] = field.metadata["long_name"]
                data[f"{field.name}/$properties"] = f"sensor,unit,{field.metadata['topic']}"
                data[f"{field.name}/sensor"] = ctx.obj["reader"].sensor.name
                data[f"{field.name}/unit"] = field.metadata["units"]
            data[f"{field.name}/{field.metadata['topic']}"] = getattr(obs, field.name)
        pub(data)

    with ctx.obj["reader"] as reader:
        publish(next(reader()), metadata=True)
        for obs in reader():
            publish(obs)



"""
NovaFitness SDS198 sensors
- messages are 10b long
- only PM100 measurements
"""

from dataclasses import dataclass, field

from pms.sensor import base
from . import sds01x

commands = sds01x.commands._replace(
    passive_read=base.Cmd(
        b"\xAA\xB4\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xFF\xFF\x02\xAB",
        b"\xAA\xCF",
        10,
    )
)


class Message(sds01x.Message):
    """Messages from NovaFitness SDS011, SDS018 and SDS021 sensors"""

    data_records = slice(1, 2)


@dataclass(frozen=False)
class ObsData(base.ObsData):

    """SDS198 observations

    time                                    measurement time [seconds since epoch]
    pm100                                   PM100 [ug/m3]
    """

    pm100: int = field(metadata=base.metadata("PM100", "ug/m3", "concentration"))

    def __format__(self, spec: str) -> str:
        if spec == "header":
            return super().__format__(spec)
        if spec == "pm":
            return f"{self.date:%F %T}: PM100 {self.pm100:.1f} ug/m3"
        if spec == "csv":
            return f"{self.time}, {self.pm100:.1f}"
        raise ValueError(  # pragma: no cover
            f"Unknown format code '{spec}' "
            f"for object of type '{__name__}.{self.__class__.__name__}'"
        )



"""
Plantower PMS5003S sensors
- messages are 32b long
- 6 size bins (as PMS5003) and HCHO concentration
"""

from dataclasses import dataclass, field

from pms.sensor import base
from . import pms3003, pmsx003


commands = pmsx003.commands


class Message(pms3003.Message):
    """Messages from Plantower PMS5003S sensors"""

    data_records = slice(13)


@dataclass(frozen=False)
class ObsData(pmsx003.ObsData):
    """Observations from Plantower PMS5003S sensors

    time                                    measurement time [seconds since epoch]
    raw01, raw25, raw10                     cf=1 PM estimates [ug/m3]
    pm01, pm25, pm10                        PM1.0, PM2.5, PM10 [ug/m3]
    n0_3, n0_5, n1_0, n2_5, n5_0, n10_0     number concentrations over X.Y um [#/cm3]
    HCHO                                    formaldehyde concentration [mg/m3]
    """

    # HCHO [mg/m3]: formaldehyde concentration (read as ug/m3, datasheet says 1/1000 mg/m3 ie ug/m3)
    HCHO: int = field(metadata=base.metadata("formaldehyde", "mg/m3", "concentration"))

    def __post_init__(self):
        """Units conversion
        nX_Y [#/cm3] read in [#/0.1L]
        HCHO [mg/m3] read in [ug/m3]
        """
        super().__post_init__()
        self.HCHO /= 1000

    def __format__(self, spec: str) -> str:
        if spec in ["header", "pm", "raw", "cf", "num"]:
            return super().__format__(spec)
        if spec == "csv":
            csv = super().__format__(spec)
            return f"{csv}, {self.HCHO:.3f}"
        if spec == "hcho":
            return f"{self.date:%F %T}: HCHO {self.HCHO:.3f} mg/m3"
        raise ValueError(  # pragma: no cover
            f"Unknown format code '{spec}' "
            f"for object of type '{__name__}.{self.__class__.__name__}'"
        )



"""
Plantower PMS1003, PMS5003, PMS7003 and PMSA003 sensors
- messages are 32b long
"""

from dataclasses import dataclass

from pms import InconsistentObservation
from pms.sensor import base
from . import pms3003


commands = base.Commands(
    passive_read=base.Cmd(b"\x42\x4D\xE2\x00\x00\x01\x71", b"\x42\x4D\x00\x1c", 32),
    passive_mode=base.Cmd(b"\x42\x4D\xE1\x00\x00\x01\x70", b"\x42\x4D\x00\x04", 8),
    active_mode=base.Cmd(b"\x42\x4D\xE1\x00\x01\x01\x71", b"\x42\x4D\x00\x1c", 32),
    sleep=base.Cmd(b"\x42\x4D\xE4\x00\x00\x01\x73", b"\x42\x4D\x00\x04", 8),
    wake=base.Cmd(b"\x42\x4D\xE4\x00\x01\x01\x74", b"\x42\x4D\x00\x1c", 32),
)


class Message(pms3003.Message):
    """Messages from Plantower PMS1003, PMS5003, PMS7003 and PMSA003 sensors"""

    data_records = slice(12)


@dataclass(frozen=False)
class ObsData(pms3003.ObsData):
    """Observations from Plantower PMS1003, PMS5003, PMS7003 and PMSA003 sensors

    time                                    measurement time [seconds since epoch]
    raw01, raw25, raw10                     cf=1 PM estimates [ug/m3]
    pm01, pm25, pm10                        PM1.0, PM2.5, PM10 [ug/m3]
    n0_3, n0_5, n1_0, n2_5, n5_0, n10_0     number concentrations over X.Y um [#/cm3]
    """

    # nX_Y [#/cm3]: number concentrations over X.Y um (read as 100*nX_Y)
    n0_3: float
    n0_5: float
    n1_0: float
    n2_5: float
    n5_0: float
    n10_0: float

    def __post_init__(self):
        """Convert from #/100cm3 to #/cm3"""
        self.n0_3 /= 100
        self.n0_5 /= 100
        self.n1_0 /= 100
        self.n2_5 /= 100
        self.n5_0 /= 100
        self.n10_0 /= 100

        if self.n0_3 == 0 and self.pm10 > 0:
            raise InconsistentObservation(
                f"inconsistent obs: PM10={self.pm10} and N0.3={self.n0_3}"
            )

    def __format__(self, spec: str) -> str:
        if spec in ["header", "pm", "raw", "cf"]:
            return super().__format__(spec)
        if spec == "csv":
            csv = super().__format__(spec)
            return f"{csv}, {self.n0_3:.2f}, {self.n0_5:.2f}, {self.n1_0:.2f}, {self.n2_5:.2f}, {self.n5_0:.2f}, {self.n10_0:.2f}"
        if spec == "num":
            return f"{self.date:%F %T}: N0.3 {self.n0_3:.2f}, N0.5 {self.n0_5:.2f}, N1.0 {self.n1_0:.2f}, N2.5 {self.n2_5:.2f}, N5.0 {self.n5_0:.2f}, N10 {self.n10_0:.2f} #/cm3"
        raise ValueError(  # pragma: no cover
            f"Unknown format code '{spec}' "
            f"for object of type '{__name__}.{self.__class__.__name__}'"
        )



"""
Honeywell HPMA115S0 sensors
- passive mode messages are 8b long
- active mode messages are 32b long
"""

from dataclasses import dataclass, field
from typing import Tuple
import struct

from pms import WrongMessageFormat, WrongMessageChecksum, SensorWarmingUp
from pms.sensor import base

commands = base.Commands(
    passive_read=base.Cmd(  # Read Particle Measuring Results
        b"\x68\x01\x04\x93", b"\x40\x05\x04", 8
    ),
    passive_mode=base.Cmd(b"\x68\x01\x20\x77", b"\xA5\xA5", 2),  # Stop Auto Send
    active_mode=base.Cmd(b"\x68\x01\x40\x57", b"\xA5\xA5", 2),  # Enable Auto Send
    sleep=base.Cmd(b"\x68\x01\x02\x95", b"\xA5\xA5", 2),  # Stop Particle Measurement
    wake=base.Cmd(b"\x68\x01\x01\x96", b"\xA5\xA5", 2),  # Start Particle Measurement
)


class Message(base.Message):
    """Messages from Honeywell HPMA115S0 sensors"""

    data_records = slice(2)

    @property
    def header(self) -> bytes:
        return self.message[:3]

    @property
    def payload(self) -> bytes:
        return self.message[3:-1]

    @property
    def checksum(self) -> int:
        return self.message[-1]

    @classmethod
    def _validate(cls, message: bytes, header: bytes, length: int) -> base.Message:

        # consistency check: bug in message singnature
        assert len(header) == 3, f"wrong header length {len(header)}"
        assert header[:1] == b"\x40", f"wrong header start {header!r}"
        assert length in [5, 8, 16], f"wrong payload length {length}"

        # validate message: recoverable errors (throw away observation)
        msg = cls(message)
        if msg.header != header:
            raise WrongMessageFormat(f"message header: {msg.header!r}")
        if len(message) != length:
            raise WrongMessageFormat(f"message length: {len(message)} != {length}")
        checksum = (0x10000 - sum(msg.header) - sum(msg.payload)) % 0x100
        if msg.checksum != checksum:
            raise WrongMessageChecksum(f"message checksum {msg.checksum} != {checksum}")
        if sum(msg.payload) == 0:
            raise SensorWarmingUp(f"message empty: warming up sensor")
        return msg

    @staticmethod
    def _unpack(message: bytes) -> Tuple[int, ...]:
        return struct.unpack(f">{len(message)//2}H", message)


@dataclass(frozen=False)
class ObsData(base.ObsData):
    """Observations from Honeywell HPMA115S0 sensors

    time                                    measurement time [seconds since epoch]
    pm25, pm10                              PM2.5, PM10 [ug/m3]
    """

    pm25: int = field(metadata=base.metadata("PM2.5", "ug/m3", "concentration"))
    pm10: int = field(metadata=base.metadata("PM10", "ug/m3", "concentration"))

    def __format__(self, spec: str) -> str:
        if spec == "header":
            return super().__format__(spec)
        if spec == "pm":
            return f"{self.date:%F %T}: PM2.5 {self.pm25:.1f}, PM10 {self.pm10:.1f} ug/m3"
        if spec == "csv":
            return f"{self.time}, {self.pm25:.1f}, {self.pm10:.1f}"
        raise ValueError(  # pragma: no cover
            f"Unknown format code '{spec}' "
            f"for object of type '{__name__}.{self.__class__.__name__}'"
        )



"""
Bosh Sensortec sensors on MCU bridge modules
- messages are 7-20b long
"""

from dataclasses import dataclass, field
from typing import Tuple, Dict
import struct

from pms import WrongMessageFormat, WrongMessageChecksum, SensorWarmingUp
from pms.sensor import base

commands = base.Commands(
    passive_read=base.Cmd(b"\xA5\x56\x01\xFC", b"\x5A\x5A\x3F\x0F", 20),
    passive_mode=base.Cmd(b"\xA5\x56\x01\xFC", b"\x5A\x5A\x3F\x0F", 20),
    active_mode=base.Cmd(b"\xA5\x56\x02\xFD", b"\x5A\x5A\x3F\x0F", 20),
    sleep=base.Cmd(b"", b"", 0),
    wake=base.Cmd(b"", b"", 0),
)


class Message(base.Message):
    """Messages from mcu680 modules with a BME680 sensor"""

    data_records = slice(7)

    @property
    def header(self) -> bytes:
        return self.message[:4]

    @property
    def payload(self) -> bytes:
        return self.message[4:-1]

    @property
    def checksum(self) -> int:
        return self.message[-1]

    @classmethod
    def _validate(cls, message: bytes, header: bytes, length: int) -> base.Message:

        # consistency check: bug in message singnature
        assert len(header) == 4, f"wrong header length {len(header)}"
        assert header[:2] == b"ZZ", f"wrong header start {header!r}"
        len_payload = header[-1]
        assert length == len_payload + 5, f"wrong payload length {length}"

        # validate message: recoverable errors (throw away observation)
        msg = cls(message)
        if msg.header != header:
            raise WrongMessageFormat(f"message header: {msg.header!r}")
        if len(message) != length:
            raise WrongMessageFormat(f"message length: {len(message)}")
        checksum = (sum(msg.header) + sum(msg.payload)) % 0x100
        if msg.checksum != checksum:
            raise WrongMessageChecksum(f"message checksum {msg.checksum} != {checksum}")
        if sum(msg.payload) == 0:
            raise SensorWarmingUp(f"message empty: warming up sensor")
        return msg

    @staticmethod
    def _unpack(message: bytes) -> Tuple[int, ...]:
        return struct.unpack(f">hHHBHLh", message)


@dataclass(frozen=False)
class ObsData(base.ObsData):
    """Observations from Plantower PMS3003 sensors

    time                                    measurement time [seconds since epoch]
    temp                                    temperature [°C]
    rhum                                    relative humidity [%]
    press                                   atmospheric pressure [hPa]
    IAQ_acc                                 IAQ accuracy flag
    IAQ                                     index of air quality [0--500]
    gas                                     gas resistance [kΩ]
    alt                                     altitude estimate [m a.s.l.]
    """

    # temp[°C],rhum[%]: temperature,relative humidity (read as 100*temp,100*rhum)
    temp: float = field(metadata=base.metadata("temperature", "°C", "degrees"))
    rhum: float = field(metadata=base.metadata("relative humidity", "%", "percentage"))
    # press[hPa]: atm. pressure (read as 24b in hPa)
    # on read press XSB(8b)|MSB(8b)
    pres: float = field(metadata=base.metadata("atmospheric pressure", "hPa", "pressure"))
    # on read press LSB(8b)
    IAQ_acc: int = field(metadata=base.metadata("IAQ acc", "1", "acc"))
    # on read IAQ_acc(4b)|IAQ(12b) packed into 16b
    IAQ: int = field(metadata=base.metadata("IAQ", "0-500", "iaq"))
    gas: int = field(metadata=base.metadata("gas resistance", "kΩ", "resistance"))
    alt: int = field(metadata=base.metadata("altitude estimate", "m(a.s.l.)", "elevation"))

    def __post_init__(self):
        """Units conversion
        temp [°C]    read in [0.01 °C]
        rhum [%]     read in [1/10 000]
        press [hPa]  read in [Pa] across 12b
        gas [kΩ]   read in [Ω]
        """
        self.temp /= 100
        self.rhum /= 100
        self.press = (int(self.pres) << 8 | self.IAQ_acc) / 100
        self.IAQ_acc = self.IAQ >> 4
        self.IAQ &= 0x0FFF
        self.gas /= 1000

    def __format__(self, spec: str) -> str:
        if spec == "header":
            return super().__format__(spec)
        if spec == "atm":
            return f"{self.date:%F %T}: Temp. {self.temp:.1f} °C, Rel.Hum. {self.rhum:.1f} %, Press {self.press:.2f} hPa"
        if spec == "bme":
            return f"{self.date:%F %T}: Temp. {self.temp:.1f} °C, Rel.Hum. {self.rhum:.1f} %, Press {self.press:.2f} hPa, {self.gas:.1f} kΩ"
        if spec == "bsec":
            return f"{self.date:%F %T}: Temp. {self.temp:.1f} °C, Rel.Hum. {self.rhum:.1f} %, Press {self.press:.2f} hPa, {self.IAQ} IAQ"
        if spec == "csv":
            return f"{self.time}, {self.temp:.1f}, {self.rhum:.1f}, {self.press:.2f}, {self.IAQ_acc}, {self.IAQ}, {self.gas:.1f}, {self.alt}"

        raise ValueError(  # pragma: no cover
            f"Unknown format code '{spec}' "
            f"for object of type '{__name__}.{self.__class__.__name__}'"
        )

    def __str__(self):
        return self.__format__("bme")



"""
Honeywell HPMA115C0 sensors
- passive mode messages are 16b long
- active mode messages are 32b long
"""

from dataclasses import dataclass, field

from pms.sensor import base
from . import hpma115s0

commands = hpma115s0.commands._replace(
    passive_read=base.Cmd(  # Read Particle Measuring Results
        b"\x68\x01\x04\x93", b"\x40\x05\x04", 16
    )
)


class Message(hpma115s0.Message):
    """Messages from Honeywell HPMA115C0 sensors"""

    data_records = slice(4)


@dataclass(frozen=False)
class ObsData(base.ObsData):
    """Observations from Honeywell HPMA115C0 sensors

    time                                    measurement time [seconds since epoch]
    pm01, pm25, pm04, pm10                  PM1.0, PM2.5, PM4.0 PM10 [ug/m3]
    """

    pm01: int = field(metadata=base.metadata("PM1", "ug/m3", "concentration"))
    pm25: int = field(metadata=base.metadata("PM2.5", "ug/m3", "concentration"))
    pm04: int = field(metadata=base.metadata("PM4", "ug/m3", "concentration"))
    pm10: int = field(metadata=base.metadata("PM10", "ug/m3", "concentration"))

    def __format__(self, spec: str) -> str:
        if spec == "header":
            return super().__format__(spec)
        if spec == "pm":
            return f"{self.date:%F %T}: PM1 {self.pm01:.1f}, PM2.5 {self.pm25:.1f}, PM4 {self.pm04:.1f}, PM10 {self.pm10:.1f} ug/m3"
        if spec == "csv":
            return (
                f"{self.time}, {self.pm01:.1f}, {self.pm25:.1f}, {self.pm04:.1f}, {self.pm10:.1f}"
            )
        raise ValueError(  # pragma: no cover
            f"Unknown format code '{spec}' "
            f"for object of type '{__name__}.{self.__class__.__name__}'"
        )



"""
Read PM sensors

NOTE:
- Sensors are read on passive mode.
- Tested on PMS3003, PMS7003, PMSA003, SDS011 and MCU680
"""

import sys
import time
from typing import Generator, Optional

from serial import Serial

from pms import logger, SensorWarning, SensorWarmingUp, InconsistentObservation
from pms.sensor import Sensor, base


class SensorReader:
    """Read sensor messages from serial port

    The sensor is woken up after opening the serial port, and put to sleep when before closing the port.
    While the serial port is open, the sensor is read in passive mode.

    PMS3003 sensors do not accept serial commands, such as wake/sleep or passive mode read.
    Valid messages are extracted from the serail buffer. Support for this sensor is experimental.
    """

    def __init__(
        self,
        sensor: str = "PMSx003",
        port: str = "/dev/ttyUSB0",
        interval: Optional[int] = None,
        samples: Optional[int] = None,
    ) -> None:
        """Configure serial port"""
        self.sensor = Sensor[sensor]
        self.serial = Serial()
        self.serial.port = port
        self.serial.baudrate = self.sensor.baud
        self.serial.timeout = 5  # max time to wake up sensor
        self.interval = interval
        self.samples = samples
        logger.debug(
            f"Reader: "
            f"request {samples if samples else '?'} obs "
            f"from {sensor} on {port} "
            f"every {interval if interval else '?'} secs"
        )

    def _cmd(self, command: str) -> bytes:
        """Write command to sensor and return answer"""

        # send command
        cmd = self.sensor.command(command)
        if cmd.command:
            self.serial.write(cmd.command)
            self.serial.flush()
        elif command.endswith("read"):
            self.serial.reset_input_buffer()

        # return full buffer
        return self.serial.read(max(cmd.answer_length, self.serial.in_waiting))

    def __enter__(self) -> "SensorReader":
        """Open serial port and sensor setup"""
        if not self.serial.is_open:
            self.serial.open()
            self.serial.reset_input_buffer()

        # wake sensor and set passive mode
        buffer = self._cmd("wake") + self._cmd("passive_mode")
        logger.debug(f"buffer length: {len(buffer)}")

        # check against sensor type derived from buffer
        if not self.sensor.check(buffer, "passive_mode"):
            logger.error(f"Sensor is not {self.sensor.name}")
            sys.exit(1)

        return self

    def __exit__(self, exception_type, exception_value, traceback) -> None:
        """Put sensor to sleep and close serial port"""
        buffer = self._cmd("sleep")
        self.serial.close()

    def __call__(self) -> Generator[base.ObsData, None, None]:
        """Passive mode reading at regular intervals"""
        while self.serial.is_open:
            try:
                buffer = self._cmd("passive_read")

                try:
                    obs = self.sensor.decode(buffer)
                except (SensorWarmingUp, InconsistentObservation) as e:
                    logger.debug(e)
                    time.sleep(5)
                except SensorWarning as e:
                    logger.debug(e)
                    self.serial.reset_input_buffer()
                else:
                    yield obs
                    if self.samples:
                        self.samples -= 1
                        if self.samples <= 0:
                            break
                    if self.interval:
                        delay = self.interval - (time.time() - obs.time)
                        if delay > 0:
                            time.sleep(delay)
            except KeyboardInterrupt:
                print()
                break



"""
NovaFitness SDS011, SDS018 and SDS021 sensors
- messages are 10b long
"""

from dataclasses import dataclass, field
from typing import Tuple
import struct

from pms import WrongMessageFormat, WrongMessageChecksum, SensorWarmingUp
from pms.sensor import base

commands = base.Commands(
    passive_read=base.Cmd(
        b"\xAA\xB4\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xFF\xFF\x02\xAB",
        b"\xAA\xC0",
        10,
    ),
    passive_mode=base.Cmd(
        b"\xAA\xB4\x02\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xFF\xFF\x02\xAB",
        b"\xAA\xC5",
        10,
    ),
    active_mode=base.Cmd(
        b"\xAA\xB4\x02\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xFF\xFF\x01\xAB",
        b"\xAA\xC5",
        10,
    ),
    sleep=base.Cmd(
        b"\xAA\xB4\x06\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xFF\xFF\x05\xAB",
        b"\xAA\xC5",
        10,
    ),
    wake=base.Cmd(
        b"\xAA\xB4\x06\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xFF\xFF\x06\xAB",
        b"\xAA\xC5",
        10,
    ),
)


class Message(base.Message):
    """Messages from NovaFitness SDS011, SDS018 and SDS021 sensors"""

    data_records = slice(2)

    @property
    def header(self) -> bytes:
        return self.message[:2]

    @property
    def payload(self) -> bytes:
        return self.message[2:-2]

    @property
    def checksum(self) -> int:
        return self.message[-2]

    @property
    def tail(self) -> int:
        return self.message[-1]

    @classmethod
    def _validate(cls, message: bytes, header: bytes, length: int) -> base.Message:

        # consistency check: bug in message singnature
        assert len(header) == 2, f"wrong header length {len(header)}"
        assert header[:1] == b"\xAA", f"wrong header start {header!r}"
        assert length == 10, f"wrong payload length {length} != 10"

        # validate message: recoverable errors (throw away observation)
        msg = cls(message)
        if msg.header != header:
            raise WrongMessageFormat(f"message header: {msg.header!r}")
        if msg.tail != 0xAB:
            raise WrongMessageFormat(f"message tail: {msg.tail:#x}")
        if len(message) != length:
            raise WrongMessageFormat(f"message length: {len(message)}")
        checksum = sum(msg.payload) % 0x100
        if msg.checksum != checksum:
            raise WrongMessageChecksum(f"message checksum {msg.checksum} != {checksum}")
        if sum(msg.payload[:-2]) == 0:
            raise SensorWarmingUp(f"message empty: warming up sensor")
        return msg

    @staticmethod
    def _unpack(message: bytes) -> Tuple[int, ...]:
        return struct.unpack(f"<{len(message)//2}H", message)


@dataclass(frozen=False)
class ObsData(base.ObsData):
    """SDS01x observations

    time                                    measurement time [seconds since epoch]
    raw25, raw10                            PM2.5*10, PM10*10 [ug/m3]
    """

    pm25: float = field(metadata=base.metadata("PM2.5", "ug/m3", "concentration"))
    pm10: float = field(metadata=base.metadata("PM10", "ug/m3", "concentration"))

    def __post_init__(self):
        """Convert from 0.1 ug/m3 to ug/m3"""
        self.pm25 /= 10
        self.pm10 /= 10

    def __format__(self, spec: str) -> str:
        if spec == "header":
            return super().__format__(spec)
        if spec == "pm":
            return f"{self.date:%F %T}: PM2.5 {self.pm25:.1f}, PM10 {self.pm10:.1f} ug/m3"
        if spec == "csv":
            return f"{self.time}, {self.pm25:.1f}, {self.pm10:.1f}"
        raise ValueError(  # pragma: no cover
            f"Unknown format code '{spec}' "
            f"for object of type '{__name__}.{self.__class__.__name__}'"
        )



"""
Plantower PMS3003 sensors
- do not support commands
- messages are 24b long
"""

from dataclasses import dataclass, field
from typing import Tuple
import struct

from pms import WrongMessageFormat, WrongMessageChecksum, SensorWarmingUp
from pms.sensor import base


commands = base.Commands(
    passive_read=base.Cmd(b"", b"\x42\x4D\x00\x14", 24),
    passive_mode=base.Cmd(b"", b"\x42\x4D\x00\x14", 24),
    active_mode=base.Cmd(b"", b"\x42\x4D\x00\x14", 24),
    sleep=base.Cmd(b"", b"\x42\x4D\x00\x14", 24),
    wake=base.Cmd(b"", b"\x42\x4D\x00\x14", 24),
)


class Message(base.Message):
    """Messages from Plantower PMS3003 sensors"""

    data_records = slice(6)

    @property
    def header(self) -> bytes:
        return self.message[:4]

    @property
    def payload(self) -> bytes:
        return self.message[4:-2]

    @property
    def checksum(self) -> int:
        return self._unpack(self.message[-2:])[0]

    @classmethod
    def _validate(cls, message: bytes, header: bytes, length: int) -> base.Message:

        # consistency check: bug in message singnature
        assert len(header) == 4, f"wrong header length {len(header)}"
        assert header[:2] == b"BM", f"wrong header start {header!r}"
        len_payload = cls._unpack(header[-2:])[0]
        assert length == 4 + len_payload, f"wrong payload length {length} != {4+len_payload}"

        # validate message: recoverable errors (throw away observation)
        msg = cls(message)
        if msg.header != header:
            raise WrongMessageFormat(f"message header: {msg.header!r}")
        if len(message) != length:
            raise WrongMessageFormat(f"message length: {len(message)}")
        checksum = sum(msg.header) + sum(msg.payload)
        if msg.checksum != checksum:
            raise WrongMessageChecksum(f"message checksum {msg.checksum} != {checksum}")
        if sum(msg.payload) == 0:
            raise SensorWarmingUp(f"message empty: warming up sensor")
        return msg

    @staticmethod
    def _unpack(message: bytes) -> Tuple[int, ...]:
        return struct.unpack(f">{len(message)//2}H", message)


@dataclass(frozen=False)
class ObsData(base.ObsData):
    """Observations from Plantower PMS3003 sensors

    time                                    measurement time [seconds since epoch]
    raw01, raw25, raw10                     cf=1 PM estimates [ug/m3]
    pm01, pm25, pm10                        PM1.0, PM2.5, PM10 [ug/m3]
    """

    raw01: int = field(repr=False)
    raw25: int = field(repr=False)
    raw10: int = field(repr=False)
    pm01: int = field(metadata=base.metadata("PM1", "ug/m3", "concentration"))
    pm25: int = field(metadata=base.metadata("PM2.5", "ug/m3", "concentration"))
    pm10: int = field(metadata=base.metadata("PM10", "ug/m3", "concentration"))

    # cfX [1]: pmX/rawX
    @property
    def cf01(self) -> float:
        return self._safe_div(self.pm01, self.raw01)

    @property
    def cf25(self) -> float:
        return self._safe_div(self.pm25, self.raw25)

    @property
    def cf10(self) -> float:
        return self._safe_div(self.pm10, self.raw10)

    @staticmethod
    def _safe_div(x: int, y: int) -> float:
        if y:
            return x / y
        if x == y == 0:  # pragma: no cover
            return 1
        return 0  # pragma: no cover

    def __format__(self, spec: str) -> str:
        if spec == "header":
            return super().__format__(spec)
        if spec == "pm":
            return f"{self.date:%F %T}: PM1 {self.pm01:.1f}, PM2.5 {self.pm25:.1f}, PM10 {self.pm10:.1f} ug/m3"
        if spec == "csv":
            return f"{self.time}, {self.raw01}, {self.raw25}, {self.raw10}, {self.pm01:.1f}, {self.pm25:.1f}, {self.pm10:.1f}"
        if spec == "cf":
            return f"{self.date:%F %T}: CF1 {self.cf01:.0%}, CF2.5 {self.cf25:.0%}, CF10 {self.cf10:.0%}"
        if spec == "raw":
            return (
                f"{self.date:%F %T}: PM1 {self.raw01}, PM2.5 {self.raw25}, PM10 {self.raw10} ug/m3"
            )
        raise ValueError(  # pragma: no cover
            f"Unknown format code '{spec}' "
            f"for object of type '{__name__}.{self.__class__.__name__}'"
        )



"""
Access supported sensors from a single object
"""

from datetime import datetime
from enum import Enum

from pms import WrongMessageFormat
from pms.sensor import base, plantower, novafitness, honeywell, senserion, bosch_sensortec


class Sensor(Enum):
    """Supported PM sensors"""

    PMSx003 = plantower.pmsx003
    PMS3003 = plantower.pms3003
    PMS5003S = plantower.pms5003s
    PMS5003ST = plantower.pms5003st
    PMS5003T = plantower.pms5003t
    SDS01x = novafitness.sds01x
    SDS198 = novafitness.sds198
    HPMA115S0 = honeywell.hpma115s0
    HPMA115C0 = honeywell.hpma115c0
    SPS30 = senserion.sps30
    MCU680 = bosch_sensortec.mcu680

    PMS1003 = PMS5003 = PMS7003 = PMSA003 = PMSx003
    G1, G3, G5, G7, G10 = PMS1003, PMS3003, PMS5003, PMS7003, PMSA003
    SDS011 = SDS018 = SDS021 = SDS01x
    BME680 = MCU680

    @property
    def Message(self) -> base.Message:
        return self.value.Message

    @property
    def Data(self) -> base.ObsData:
        return self.value.ObsData

    @property
    def Commands(self) -> base.Commands:
        return self.value.commands

    @property
    def baud(self) -> int:  # pragma: no cover
        return 115_200 if self.name == "SPS30" else 9600

    @staticmethod
    def now() -> int:  # pragma: no cover
        """current time as seconds since epoch"""
        return int(datetime.now().timestamp())

    def command(self, cmd: str) -> base.Cmd:
        """Serial command for sensor"""
        return getattr(self.Commands, cmd)

    def check(self, buffer: bytes, command: str) -> bool:
        """Validate buffer contents"""
        try:
            self.Message.decode(buffer, self.command(command))
        except WrongMessageFormat:
            return False
        else:
            return True

    def decode(self, buffer: bytes, *, time: int = None) -> base.ObsData:
        """Exract observations from serial buffer"""
        if not time:  # pragma: no cover
            time = self.now()

        data = self.Message.decode(buffer, self.Commands.passive_read)
        return self.Data(time, *data)  # type: ignore



"""Additional commands for Honeywell sensors

HPM Series, Particulate Matter Sensors, 32322550 Issue F
https://sensing.honeywell.com/honeywell-sensing-particulate-hpm-series-datasheet-32322550
"""

from pms.sensor.base import Cmd


def read_cf() -> Cmd:
    """Read Customer Adjustment Coefficient

    HPM Series, Table 4 and Table 6
    """
    return Cmd(b"\x68\x01\x10\x87", b"\x40\x02\x10", 5)


def write_cf(cf: int = 0) -> Cmd:
    """Set Customer Adjustment Coefficient

    HPM Series, Table 4 and Table 6
    cf: 30 ~ 200 (Default, 100)
    """

    assert 30 <= cf <= 200, f"out of range: 30 <= {cf} <= 200"
    return Cmd(bytes.fromhex(f"680208{cf:02X}{(0xff8e-cf)%0x100:02X}"), b"\xA5\xA5", 2)



"""
Plantower PMS5003T sensors
- messages are 32b long
- only 4 size bins
"""

from dataclasses import dataclass, field
from typing import Tuple
import struct

from pms import InconsistentObservation
from pms.sensor import base
from . import pms3003, pmsx003


commands = pmsx003.commands


class Message(pms3003.Message):
    """Messages from Plantower PMS5003T sensors"""

    data_records = slice(12)

    @staticmethod
    def _unpack(message: bytes) -> Tuple[int, ...]:
        if len(message) == 26:
            # 11th record is signed (temp)
            return struct.unpack(">10Hh2H", message)
        else:
            return pms3003.Message._unpack(message)


@dataclass(frozen=False)
class ObsData(pms3003.ObsData):
    """Observations from Plantower PMS5003T sensors

    time                                    measurement time [seconds since epoch]
    raw01, raw25, raw10                     cf=1 PM estimates [ug/m3]
    pm01, pm25, pm10                        PM1.0, PM2.5, PM10 [ug/m3]
    n0_3, n0_5, n1_0, n2_5                  number concentrations over X.Y um [#/cm3]
    temp                                    temperature [°C]
    rhum                                    relative humidity [%]
    """

    # nX_Y [#/cm3]: number concentrations over X.Y um (read as 100*nX_Y)
    n0_3: float
    n0_5: float
    n1_0: float
    n2_5: float
    # temp[°C],rhum[%]: temperature,relative humidity (read as 10*temp,10*rhum)
    temp: float = field(metadata=base.metadata("temperature", "°C", "degrees"))
    rhum: float = field(metadata=base.metadata("relative humidity", "%", "percentage"))

    def __post_init__(self):
        """Units conversion
        nX_Y [#/cm3] read in [#/0.1L]
        temp [°C]    read in [0.1 °C]
        rhum [%]     read in [1/1000]
        """
        self.n0_3 /= 100
        self.n0_5 /= 100
        self.n1_0 /= 100
        self.n2_5 /= 100
        self.temp /= 10
        self.rhum /= 10

        if self.n0_3 == 0 and self.pm10 > 0:
            raise InconsistentObservation(
                f"inconsistent obs: PM10={self.pm10} and N0.3={self.n0_3}"
            )

    def __format__(self, spec: str) -> str:
        if spec in ["header", "pm", "raw", "cf"]:
            return super().__format__(spec)
        if spec == "csv":
            pm = super().__format__(spec)
            return f"{pm}, {self.n0_3:.2f}, {self.n0_5:.2f}, {self.n1_0:.2f}, {self.n2_5:.2f}, {self.temp:.1f}, {self.rhum:.1f}"
        if spec == "num":
            return f"{self.date:%F %T}: N0.3 {self.n0_3:.2f}, N0.5 {self.n0_5:.2f}, N1.0 {self.n1_0:.2f}, N2.5 {self.n2_5:.2f} #/cm3"
        if spec == "atm":
            return f"{self.date:%F %T}: Temp. {self.temp:.1f} °C, Rel.Hum. {self.rhum:.1f} %"
        raise ValueError(  # pragma: no cover
            f"Unknown format code '{spec}' "
            f"for object of type '{__name__}.{self.__class__.__name__}'"
        )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pms',
 'pms.sensor',
 'pms.sensor.bosch_sensortec',
 'pms.sensor.honeywell',
 'pms.sensor.novafitness',
 'pms.sensor.plantower',
 'pms.sensor.senserion',
 'pms.service']

package_data = \
{'': ['*']}

install_requires = \
['mypy-extensions>=0.4.0', 'pyserial>=3.0', 'typer>=0.3.0']

extras_require = \
{':python_version >= "3.6" and python_version < "4.0"': ['dataclasses>=0.6'],
 'influxdb': ['influxdb>=5.2.0'],
 'mqtt': ['paho-mqtt>=1.4.0']}

entry_points = \
{'console_scripts': ['pms = pms.cli:main']}

setup_kwargs = {
    'name': 'pypms',
    'version': '0.3.1',
    'description': 'Data acquisition and logging tool for PM sensors with UART interface',
    'long_description': "# Serial Air Quality Sensors\n\nTools for reading Air Quality Sensors with serial (UART) interface, data acquisition and logging.\n\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/pypms)](https://pypi.org/project/pypms)\n[![PyPI](https://img.shields.io/pypi/v/pypms)](https://pypi.org/project/pypms)\n[![Build Status](https://travis-ci.com/avaldebe/PyPMS.svg?branch=master)](https://travis-ci.com/avaldebe/PyPMS)\n[![GitHub issues](https://img.shields.io/github/issues/avaldebe/PyPMS)](https://github.com/avaldebe/PyPMS/issues)\n[![GitHub license](https://img.shields.io/github/license/avaldebe/PyPMS)](https://github.com/avaldebe/PyPMS/blob/master/LICENSE)\n[![Checked with mypy](http://www.mypy-lang.org/static/mypy_badge.svg)](http://mypy-lang.org/)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![DOI](https://zenodo.org/badge/203110737.svg)](https://zenodo.org/badge/latestdoi/203110737)\n\n## Command Line Interface\n\n```man\nUsage: pms [OPTIONS] COMMAND [ARGS]...\n\n  Read serial sensor\n\nOptions:\n  --sensor-model -m [PMSx003|PMS3003|PMS5003S|PMS5003ST|PMS5003T|SDS01x|SDS198|HPMA115S0|HPMA115C0|SPS30|MCU680]\n                                  sensor model  [default: PMSx003]\n  -s, --serial-port TEXT          serial port  [default: /dev/ttyUSB0]\n  -i, --interval INTEGER          seconds to wait between updates  [default:\n                                  60]\n  -n, --samples INTEGER           stop after N samples\n  --debug                         print DEBUG/logging messages  [default:\n                                  False]\n\n  --install-completion [bash|zsh|fish|powershell|pwsh]\n                                  Install completion for the specified shell.\n  --show-completion [bash|zsh|fish|powershell|pwsh]\n                                  Show completion for the specified shell, to\n                                  copy it or customize the installation.\n\n  --help                          Show this message and exit.\n\nCommands:\n  bridge    Bridge between MQTT and InfluxDB servers\n  csv       Read sensor and print measurements\n  influxdb  Read sensor and push PM measurements to an InfluxDB server\n  mqtt      Read sensor and push PM measurements to a MQTT server\n  serial    Read sensor and print measurements\n```\n\nFor details on a particular command and their options\n\n```bash\npms COMMAND --help\n```\n\n## Installation\n\n```bash\n# basic installation with pip\npython3 -m pip install pypms\n\n# or with pipx\npipx install pypms\n```\n\nWill allow you yo access to sensors via serial port (`pms serial`),\nand save observations to a csv file (`pms csv`).\n\nAdditional packages are required for pushing observations to an mqtt server\n(`pms mqtt`), to an influxdb server (`pms influxdb`), or provide a bridge\nbetween mqtt and influxdb servers (`pms bridge`).\n\n```bash\n# full installation with pip\npython3 -m pip install pypms[mqtt,influxdb]\n\n# or with pipx\npipx install pypms[mqtt,influxdb]\n```\n\n## Particulate Matter Sensors\n\n| Sensor            | `--sensor-model` |  PM1  | PM2.5 |  PM4  | PM10  | size bins | Other                  | Tested Works | Doesn't Work | Not Tested | Datasheet                     | Notes                |\n| ----------------- | ---------------- | :---: | :---: | :---: | :---: | :-------: | ---------------------- | :----------: | :----------: | :--------: | ----------------------------- | -------------------- |\n| [Plantower][]     |\n| PMS1003 (aka G1)  | `PMSx003`        |   X   |   X   |       |   X   |     6     |                        |              |              |     X      | [en][g1_aqmd],  [cn][g1_lcsc] |\n| PMS3003 (aka G3)  | `PMS3003`        |   X   |   X   |       |   X   |           |                        |      X       |              |            | [en][g3_aqmon], [cn][g3_lcsc] | No passive mode read |\n| PMS5003 (aka G5)  | `PMSx003`        |   X   |   X   |       |   X   |     6     |                        |              |              |     X      | [en][g5_aqmd],  [cn][g5_lcsc] |\n| PMS5003S          | `PMS5003S`       |   X   |   X   |       |   X   |     6     | HCHO concentration     |              |              |     X      |\n| PMS5003ST         | `PMS5003ST`      |   X   |   X   |       |   X   |     6     | HCHO, temp. & rel.hum. |              |              |     X      |\n| PMS5003T          | `PMS5003T`       |   X   |   X   |       |   X   |     4     | temp. & rel.hum.       |              |              |     X      |\n| PMS7003 (aka G7)  | `PMSx003`        |   X   |   X   |       |   X   |     6     |                        |      X       |              |            | [cn][g7_lcsc]                 |\n| PMSA003 (aka G10) | `PMSx003`        |   X   |   X   |       |   X   |     6     |                        |      X       |              |            | [cn][gA_lcsc]                 |\n| [NovaFitness][]   |\n| SDS011            | `SDS01x`         |       |   X   |       |   X   |           |                        |      X       |              |            | [en][SDS011]                  |\n| SDS018            | `SDS01x`         |       |   X   |       |   X   |           |                        |              |              |     X      | [en][SDS018]                  |\n| SDS021            | `SDS01x`         |       |   X   |       |   X   |           |                        |              |              |     X      | [en][SDS021]                  |\n| SDS198            | `SDS198`         |       |       |       |       |           | PM100                  |              |              |     X      | [en][SDS198]                  |\n| [Honeywell][]     |\n| HPMA115S0         | `HPMA115S0`      |       |   X   |       |   X   |           |                        |              |              |     X      | [en][HPMA115]                 |\n| HPMA115C0         | `HPMA115C0`      |   X   |   X   |   X   |   X   |           |                        |              |              |     X      | [en][HPMA115]                 |\n| [Senserion][]     |\n| SPS30             | `SPS30`          |   X   |   X   |   X   |   X   |     5     | typical particle size  |              |              |     X      | [en][SPS30]                   | UART 115200 8N1      |\n\n[plantower]: http://www.plantower.com/\n[g1_aqmd]:    http://www.aqmd.gov/docs/default-source/aq-spec/resources-page/plantower-pms1003-manual_v2-5.pdf?sfvrsn=2\n[g5_aqmd]:    http://www.aqmd.gov/docs/default-source/aq-spec/resources-page/plantower-pms5003-manual_v2-3.pdf?sfvrsn=2\n[g3_aqmon]:   https://github.com/avaldebe/AQmon/raw/master/Documents/PMS3003_LOGOELE.pdf\n[g5_aqmon]:   https://github.com/avaldebe/AQmon/raw/master/Documents/PMS5003_LOGOELE.pdf\n[g1_lcsc]:    https://datasheet.lcsc.com/szlcsc/PMS1003_C89289.pdf\n[g3_lcsc]:    https://datasheet.lcsc.com/szlcsc/PMS3003_C87024.pdf\n[g5_lcsc]:    https://datasheet.lcsc.com/szlcsc/PMS5003_C91431.pdf\n[g7_lcsc]:    https://datasheet.lcsc.com/szlcsc/PMS7003_C84815.pdf\n[gA_lcsc]:    https://datasheet.lcsc.com/szlcsc/PMSA003-A_C132744.pdf\n\n[NovaFitness]: http://inovafitness.com/en/a/index.html\n[SDS011]: https://www-sd-nf.oss-cn-beijing.aliyuncs.com/官网下载/SDS011%20laser%20PM2.5%20sensor%20specification-V1.4.pdf\n[SDS018]: https://www-sd-nf.oss-cn-beijing.aliyuncs.com/官网下载/SDS018%20Laser%20PM2.5%20Product%20Spec%20V1.5.pdf\n[SDS021]: https://cdn.sparkfun.com/assets/parts/1/2/2/7/5/SDS021_laser_PM2.5_sensor_specification-V1.0.pdf\n[SDS198]: https://www-sd-nf.oss-cn-beijing.aliyuncs.com/官网下载/SDS198%20laser%20PM100%20sensor%20specification-V1.2.pdf\n\n[Honeywell]: https://sensing.honeywell.com/sensors/particle-sensors/hpm-series\n[HPMA115]: https://sensing.honeywell.com/honeywell-sensing-particulate-hpm-series-datasheet-32322550\n\n[Senserion]: https://www.sensirion.com/en/environmental-sensors/particulate-matter-sensors-pm25/\n[SPS30]: https://www.sensirion.com/fileadmin/user_upload/customers/sensirion/Dokumente/9.6_Particulate_Matter/Datasheets/Sensirion_PM_Sensors_SPS30_Datasheet.pdf\n\n## Other Sensors\n\n- [MCU680][]:\n  chinese module with a [BME680][] sensor, [STM32F051K8][] mirocontroller (μC) and 3.3V low-dropout regulator (LDO).\n  The μC acts as I2C/UART bridge, providing outputs from the [closed source integration library][BSEC].\n\n[MCU680]:   /docs/sensors/mcu680.md\n[BME680]:   https://www.bosch-sensortec.com/media/boschsensortec/downloads/datasheets/bst-bme680-ds001.pdf\n[STM32F051K8]:  https://www.st.com/en/microcontrollers-microprocessors/stm32f051k8.html\n[BSEC]:     https://www.bosch-sensortec.com/software-tools/software/bsec/\n\n## Want More Sensors\n\nFor more Air Quality sensors [open an issue][issue].\n\n[issue]: https://github.com/avaldebe/PyPMS/issues\n\n## Changelog\n\n- 0.3.1\n  - fix influxdb default tags\n- 0.3.0\n  - option for a fix number of samples\n  - PMSx003 consistency check after sleep/wake\n- 0.2.*\n  - widen project scope from PM sensors to AQ sensors in general\n  - support [BME680][] sensor ([MCU680][] module)\n- 0.1.*\n  - widen project scope beyond [Plantower][] PM sensors\n  - support [NovaFitness][], [Honeywell][] and [Senserion][]\xa0PM sensors\n  - cli for logging to csv file, InfluxDB server or MQTT server\n",
    'author': 'Alvaro Valdebenito',
    'author_email': 'avaldebe@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/avaldebe/PyPMS',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import json
from dataclasses import fields
from typing import Dict, Callable

from typer import Context, Option, style, colors, echo, Abort
from mypy_extensions import NamedArg

try:
    from influxdb import InfluxDBClient as client
except ModuleNotFoundError:  # pragma: no cover
    client = None  # type: ignore


def __missing_influxdb():  # pragma: no cover
    name = style(__name__, fg=colors.GREEN, bold=True)
    package = style("pypms", fg=colors.GREEN, bold=True)
    module = style("influxdb", fg=colors.RED, bold=True)
    extra = style("influxdb", fg=colors.RED, bold=True)
    pip = style("python3 -m pip instal --upgrade", fg=colors.GREEN)
    pipx = style("pipx inject", fg=colors.GREEN)
    echo(
        f"""
{name} provides additional functionality to {package}.
This functionality requires the {module} module, which is not installed.
You can install this additional dependency with
\t{pip} {package}[{extra}]
Or, if you installed {package} with pipx
\t{pipx} {package} {module}
"""
    )
    raise Abort()


def client_pub(
    *, host: str, port: int, username: str, password: str, db_name: str
) -> Callable[
    [NamedArg(int, "time"), NamedArg(Dict[str, str], "tags"), NamedArg(Dict[str, float], "data")],
    None,
]:  # pragma: no cover
    if client is None:
        __missing_influxdb()
    c = client(host, port, username, password, None)
    databases = c.get_list_database()
    if len(list(filter(lambda x: x["name"] == db_name, databases))) == 0:
        c.create_database(db_name)
    c.switch_database(db_name)

    def pub(*, time: int, tags: Dict[str, str], data: Dict[str, float]) -> None:
        c.write_points(
            [
                {"measurement": k, "tags": tags, "time": time, "fields": {"value": v}}
                for k, v in data.items()
            ],
            time_precision="s",
        )

    return pub


def influxdb(
    ctx: Context,
    host: str = Option("influxdb", "--db-host", help="database server"),
    port: int = Option(8086, "--db-port", help="server port"),
    user: str = Option("root", "--db-user", help="server username"),
    word: str = Option("root", "--db-pass", help="server password"),
    name: str = Option("homie", "--db-name", help="database name"),
    jtag: str = Option(json.dumps({"location": "test"}), "--tags", help="measurement tags"),
):
    """Read sensor and push PM measurements to an InfluxDB server"""
    pub = client_pub(host=host, port=port, username=user, password=word, db_name=name)
    tags = json.loads(jtag.replace("'", '"'))

    with ctx.obj["reader"] as reader:
        for obs in reader():
            data = {field.name: getattr(obs, field.name) for field in fields(obs) if field.metadata}
            pub(time=obs.time, tags=tags, data=data)



"""
Plantower PMS5003ST sensors
- messages are 40b long
- 6 size bins (as PMS5003). HCHO concentration, temperature and relative humidity
"""

from dataclasses import dataclass, field
from typing import Tuple
import struct

from pms.sensor import base
from . import pms3003, pmsx003, pms5003s


commands = base.Commands(
    passive_read=base.Cmd(pmsx003.commands.passive_read.command, b"\x42\x4D\x00\x24", 40),
    passive_mode=pmsx003.commands.passive_mode,
    active_mode=base.Cmd(pmsx003.commands.active_mode.command, b"\x42\x4D\x00\x24", 40),
    sleep=pmsx003.commands.sleep,
    wake=base.Cmd(pmsx003.commands.wake.command, b"\x42\x4D\x00\x24", 40),
)


class Message(pms3003.Message):
    """Messages from Plantower PMS5003ST sensors"""

    data_records = slice(15)

    @staticmethod
    def _unpack(message: bytes) -> Tuple[int, ...]:
        if len(message) == 34:
            # 14th record is signed (temp)
            return struct.unpack(f">13Hh3H", message)
        else:
            return pms3003.Message._unpack(message)


@dataclass(frozen=False)
class ObsData(pms5003s.ObsData):
    """Observations from Plantower PMS5003ST sensors

    time                                    measurement time [seconds since epoch]
    raw01, raw25, raw10                     cf=1 PM estimates [ug/m3]
    pm01, pm25, pm10                        PM1.0, PM2.5, PM10 [ug/m3]
    n0_3, n0_5, n1_0, n2_5, n5_0, n10_0     number concentrations over X.Y um [#/cm3]
    HCHO                                    formaldehyde concentration [mg/m3]
    temp                                    temperature [°C]
    rhum                                    relative humidity [%]
    """

    # temp[°C],rhum[%]: temperature,relative humidity (read as 10*temp,10*rhum)
    temp: float = field(metadata=base.metadata("temperature", "°C", "degrees"))
    rhum: float = field(metadata=base.metadata("relative humidity", "%", "percentage"))

    def __post_init__(self):
        """Units conversion
        nX_Y [#/cm3] read in [#/0.1L]
        HCHO [mg/m3] read in [ug/m3]
        temp [°C]    read in [0.1 °C]
        rhum [%]     read in [1/1000]
        """
        super().__post_init__()
        self.temp /= 10
        self.rhum /= 10

    def __format__(self, spec: str) -> str:
        if spec in ["header", "pm", "raw", "cf", "num", "hcho"]:
            return super().__format__(spec)
        if spec == "csv":
            csv = super().__format__(spec)
            return f"{csv}, {self.temp:.1f}, {self.rhum:.1f}"
        if spec == "atm":
            return f"{self.date:%F %T}: Temp. {self.temp:.1f} °C, Rel.Hum. {self.rhum:.1f} %"
        raise ValueError(  # pragma: no cover
            f"Unknown format code '{spec}' "
            f"for object of type '{__name__}.{self.__class__.__name__}'"
        )



"""Additional commands for NovaFitness sensors

Laser Dust Sensor Control Protocol V1.3
https://learn.watterott.com/sensors/sds011/sds011_protocol.pdf
"""


from pms.sensor.base import Cmd


def _msg(cmd: int, payload: str, device: int = 0xFFFF) -> bytes:
    assert 0 <= cmd <= 0xFF, f"command id out of range: 0 <= {cmd} <= 0xFF"
    assert 0 <= device <= 0xFFFF, f"device id out of range: 0 <= {device} <= 0xFFFF"
    checksum = sum(bytes.fromhex(payload)) + device // 0x100 + device % 0x100
    return bytes.fromhex(f"AA{cmd:02X}{payload}{device:04X}{checksum%0x100:02X}AB")


def write_id(new_id: int, device: int = 0xFFFF) -> Cmd:
    """Protocol V1.3, section 3) Set Device ID

    The setting is still effective after power off [Factory default has set a unique ID]
    """
    return Cmd(_msg(0xB4, f"0500000000000000000000{new_id:02X}", device), b"\xAA\xC0", 10)


def work_period(minutes: int = 0, device: int = 0xFFFF) -> Cmd:
    """Protocol V1.3, section 5) Set working period

    The setting is still effective after power off [factory default is continuous measurement]
    The sensor works periodically and reports the latest data.

    0 minute: continuous mode (default), i.e. report every 1 second
    1-30 minutes: sample for 30 secors and sleep the rest of the period
    """

    assert 0 <= minutes <= 30, f"minutes out of range: 0 <= {minutes} <= 30"
    return Cmd(_msg(0xB4, f"0801{minutes:02X}00000000000000000000", device), b"\xAA\xC0", 10)


def firmware_version(device: int = 0xFFFF) -> Cmd:
    """Protocol V1.3, 6) Check firmware version"""
    return Cmd(_msg(0xB4, "07000000000000000000000000", device), b"\xAA\xC0", 10)
